@@ -1,13 +1,31 @@
-from typing import Any, Literal, Optional, Tuple, TYPE_CHECKING
+from typing import (
+    Any,
+    Iterable,
+    Literal,
+    Optional,
+    Tuple,
+    TYPE_CHECKING,
+    Union,
+    overload,
+)
 from rqdb.errors import DBError
+from rqdb.explain import (
+    ExplainQueryPlan,
+    async_write_explain_query_plan,
+    format_explain_query_plan_result,
+    parse_explain_query_plan,
+    print_explain_query_plan_result,
+    write_explain_query_plan,
+)
 from rqdb.logging import log
 from rqdb.result import BulkResult, ResultItem, ResultItemCursor
 from rqdb.preprocessing import get_sql_command, clean_nulls
+import inspect
 import time
 import secrets
 
 if TYPE_CHECKING:
-    from rqdb.async_connection import AsyncConnection
+    from rqdb.async_connection import AsyncConnection, SyncWritableIO, AsyncWritableIO
 
 
 class AsyncCursor:
@@ -22,7 +40,7 @@ class AsyncCursor:
         self.connection = connection
         """The underlying connection to the rqlite cluster."""
 
-        self.read_consistency = read_consistency
+        self.read_consistency: Literal["none", "weak", "strong"] = read_consistency
         """The read consistency to use when executing queries."""
 
         self.freshness = freshness
@@ -47,7 +65,7 @@ class AsyncCursor:
     async def execute(
         self,
         operation: str,
-        parameters: Optional[tuple] = None,
+        parameters: Optional[Iterable[Any]] = None,
         raise_on_error: bool = True,
         read_consistency: Optional[Literal["none", "weak", "strong"]] = None,
         freshness: Optional[str] = None,
@@ -58,7 +76,7 @@ class AsyncCursor:
 
         Args:
             operation (str): The query to execute.
-            parameters (tuple): The parameters to pass to the query.
+            parameters (iterable): The parameters to pass to the query.
             raise_on_error (bool): If True, raise an error if the query fails. If
                 False, you can check the result item's error property to see if
                 the query failed.
@@ -85,12 +103,12 @@ class AsyncCursor:
 
         command = get_sql_command(operation)
         cleaned_query, parameters = clean_nulls(operation, parameters)
-        is_read = command == "SELECT"
+        is_read = command in ("SELECT", "EXPLAIN")
         request_id = secrets.token_hex(4)
 
         if is_read:
 
-            def msg_supplier(max_length: Optional[int]) -> str:
+            def msg_supplier_1(max_length: Optional[int]) -> str:
                 abridged_query = operation
                 if max_length is not None and len(abridged_query) > max_length:
                     abridged_query = abridged_query[:max_length] + "..."
@@ -105,7 +123,7 @@ class AsyncCursor:
 
                 return f"  [RQLITE {command} @ {read_consistency}{freshness_str} {{{request_id}}}] - {repr(abridged_query)}; {abridged_parameters}"
 
-            log(self.connection.log_config.read_start, msg_supplier)
+            log(self.connection.log_config.read_start, msg_supplier_1)
 
             path = f"/db/query?level={read_consistency}"
             if self.read_consistency == "none":
@@ -194,8 +212,8 @@ class AsyncCursor:
 
     async def executemany2(
         self,
-        operations: Tuple[str],
-        seq_of_parameters: Optional[Tuple[Tuple[Any]]] = None,
+        operations: Iterable[str],
+        seq_of_parameters: Optional[Iterable[Iterable[Any]]] = None,
         transaction: bool = True,
         raise_on_error: bool = True,
     ) -> BulkResult:
@@ -209,8 +227,8 @@ class AsyncCursor:
         as if they are updates, i.e., no result rows will be returned.
 
         Args:
-            operations (Tuple[str]): The operations to execute.
-            seq_of_parameters (Tuple[Tuple[Any]]): The parameters to pass to each operation.
+            operations (iterable[str]): The operations to execute.
+            seq_of_parameters (iterable[iterable[Any]]): The parameters to pass to each operation.
             transaction (bool): If True, execute the operations within a transaction.
             raise_on_error (bool): If True, raise an error if any of the operations fail. If
                 False, you can check the result item's error property to see if the
@@ -225,30 +243,51 @@ class AsyncCursor:
         if seq_of_parameters is None:
             seq_of_parameters = tuple(tuple() for _ in operations)
 
-        if len(operations) != len(seq_of_parameters):
-            raise ValueError(
-                f"Number of operations ({len(operations)}) does not match number of parameters ({len(seq_of_parameters)})"
-            )
-
         path = "/db/execute"
         if transaction:
             path += "?transaction"
 
         cleaned_request = []
-        for operation, parameters in zip(operations, seq_of_parameters):
+
+        operations_iter = iter(operations)
+        seq_of_parameters_iter = iter(seq_of_parameters)
+
+        next_operation: str = ""
+        next_parameters: Iterable[Any] = tuple()
+
+        while True:
+            try:
+                next_operation = next(operations_iter)
+            except StopIteration:
+                operations_iter = None
+            try:
+                next_parameters = next(seq_of_parameters_iter)
+            except StopIteration:
+                seq_of_parameters_iter = None
+
+            if (operations_iter is None) is not (seq_of_parameters_iter is None):
+                raise ValueError(
+                    "operations and seq_of_parameters must be the same length"
+                )
+
+            if operations_iter is None or seq_of_parameters_iter is None:
+                break
+
+            operation: str = next_operation
+            parameters: Iterable[Any] = next_parameters
             cleaned_query, parameters = clean_nulls(operation, parameters)
             cleaned_request.append([cleaned_query, *parameters])
 
         request_id = secrets.token_hex(4)
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_1(max_length: Optional[int]) -> str:
             abridged_request = repr(cleaned_request)
             if max_length is not None and len(abridged_request) > max_length:
                 abridged_request = abridged_request[:max_length] + "..."
 
             return f"  [RQLITE BULK {path} {{{request_id}}}] - {abridged_request}"
 
-        log(self.connection.log_config.write_start, msg_supplier)
+        log(self.connection.log_config.write_start, msg_supplier_1)
 
         request_started_at = time.perf_counter()
         response = await self.connection.fetch_response(
@@ -261,7 +300,7 @@ class AsyncCursor:
         await response.__aexit__(None, None, None)
         request_time = time.perf_counter() - request_started_at
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_2(max_length: Optional[int]) -> str:
             abridged_payload = repr(payload)
             if max_length is not None and len(abridged_payload) > max_length:
                 abridged_payload = abridged_payload[:max_length] + "..."
@@ -270,7 +309,7 @@ class AsyncCursor:
 
         log(
             self.connection.log_config.write_response,
-            msg_supplier,
+            msg_supplier_2,
         )
 
         result = BulkResult.parse(payload)
@@ -281,7 +320,7 @@ class AsyncCursor:
 
     async def executemany3(
         self,
-        operation_and_parameters: Tuple[Tuple[str, Tuple[Any]]],
+        operation_and_parameters: Iterable[Tuple[str, Iterable[Any]]],
         transaction: bool = True,
         raise_on_error: bool = True,
     ) -> BulkResult:
@@ -319,14 +358,14 @@ class AsyncCursor:
 
         request_id = secrets.token_hex(4)
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_1(max_length: Optional[int]) -> str:
             abridged_request = repr(cleaned_request)
             if max_length is not None and len(abridged_request) > max_length:
                 abridged_request = abridged_request[:max_length] + "..."
 
             return f"  [RQLITE BULK {path} {{{request_id}}}] - {abridged_request}"
 
-        log(self.connection.log_config.write_start, msg_supplier)
+        log(self.connection.log_config.write_start, msg_supplier_1)
 
         request_started_at = time.perf_counter()
         response = await self.connection.fetch_response(
@@ -339,7 +378,7 @@ class AsyncCursor:
         await response.__aexit__(None, None, None)
         request_time = time.perf_counter() - request_started_at
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_2(max_length: Optional[int]) -> str:
             abridged_payload = repr(payload)
             if max_length is not None and len(abridged_payload) > max_length:
                 abridged_payload = abridged_payload[:max_length] + "..."
@@ -348,7 +387,7 @@ class AsyncCursor:
 
         log(
             self.connection.log_config.write_response,
-            msg_supplier,
+            msg_supplier_2,
         )
 
         result = BulkResult.parse(payload)
@@ -357,6 +396,182 @@ class AsyncCursor:
 
         return result
 
+    @overload
+    async def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Literal["str"] = "str",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+    ) -> str:
+        ...
+
+    @overload
+    async def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Literal["print"],
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+    ) -> None:
+        ...
+
+    @overload
+    async def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Literal["plan"],
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+    ) -> ExplainQueryPlan:
+        ...
+
+    @overload
+    async def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: "SyncWritableIO",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+        sync: Optional[Literal[True]] = None,
+    ) -> None:
+        ...
+
+    @overload
+    async def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: "AsyncWritableIO",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+        sync: Optional[Literal[False]] = None,
+    ) -> None:
+        ...
+
+    async def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Union[
+            Literal["str", "print", "plan"], "SyncWritableIO", "AsyncWritableIO"
+        ] = "str",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+        sync: Optional[bool] = None,
+    ) -> Union[ExplainQueryPlan, str, None]:
+        """Accepts any query; if it is not prefixed with EXPLAIN then
+        it will be prefixed with EXPLAIN QUERY PLAN. The reuslt will
+        then be parsed into the corresponding tree structure and either
+        written to the designated output or returned.
+
+        This always raises if there is an error, and can only operate
+        at none/weak level. If the read consistency is not specified
+        and the cursor read consistency is strong, weak consistency
+        is used instead.
+
+        Args:
+            operation (str): The query to execute.
+            out (Union[Literal["str", "print", "plan"], SyncWritableIO, AsyncWritableIO]):
+                The output to write the explain query plan to. If "str",
+                return the explain query plan as a string. If "print",
+                print the explain query plan to stdout. If "plan", return
+                the explain query plan as a tree structure. If a writable
+                stream, write the explain query plan to the stream.
+
+                Async vs sync is detected via if the function is coroutine or
+                not unless specified via "sync"
+            parameters (Optional[Iterable[Any]]): The parameters to pass to the query.
+                Parameters must be specified if the operation is a parameterized query,
+                though the values generally only need to be of the correct shape for
+                the appropriate plan to be returned
+            read_consistency ("none", "weak", None): The read consistency to use,
+                None for a random node, weak for the current leader, or None for the
+                cursor default (downgrading strong to weak as query plans are not
+                necessarily deterministic)
+            freshness (Optional[str]): The freshness to use when executing
+                none read consistency queries. If None, use the default freshness
+                for this cursor.
+            indent (int): The number of characters to indent each level when
+                formatting the plan. Ignored for "plan" output.
+            include_raw (bool): If True, include the ids that made each row in the
+                formatted output. Ignored for "plan" output.
+            sync (Optional[bool]): Ignored unless the output is a writable stream.
+                If None, we check if the `write` function is a coroutine function,
+                i.e., is defined via `async def`. If True, we assume it's synchronous,
+                and if False, we assume it's asynchronous. Generally you don't need
+                to specify this unless the stream returns an awaitable but isn't a
+                typical coroutine function.
+
+        Returns:
+            Depends on the value of out. If "str", return the explain query plan as a
+            string. If "print", print the explain query plan to stdout. If "plan",
+            return the explain query plan as a tree structure. If a writable stream,
+            write the explain query plan to the bytes stream and return None
+        """
+        command = get_sql_command(operation)
+        if command != "EXPLAIN":
+            operation = f"EXPLAIN QUERY PLAN {operation}"
+
+        if read_consistency is None:
+            read_consistency = (
+                self.read_consistency if self.read_consistency != "strong" else "weak"
+            )
+
+        result = await self.execute(
+            operation,
+            parameters,
+            read_consistency=read_consistency,
+            freshness=freshness,
+        )
+        if out == "plan":
+            return parse_explain_query_plan(result)
+        elif out == "str":
+            return format_explain_query_plan_result(
+                result, indent=indent, include_raw=include_raw
+            )
+        elif out == "print":
+            print_explain_query_plan_result(
+                result, indent=indent, include_raw=include_raw
+            )
+            return
+        else:
+            assert not isinstance(out, str), f"unrecognized out value: {out}"
+
+            plan = parse_explain_query_plan(result)
+
+            if sync is False or (
+                sync is None and inspect.iscoroutinefunction(out.write)
+            ):
+                await async_write_explain_query_plan(
+                    plan, out, indent=indent, include_raw=include_raw
+                )
+            else:
+                write_explain_query_plan(
+                    plan, out, indent=indent, include_raw=include_raw
+                )
+            return
+
     def fetchone(self) -> Optional[list]:
         """Fetches the next row from the cursor. If there are no more rows,
         returns None.
