@@ -1,5 +1,21 @@
-from typing import Any, Literal, Optional, Tuple, TYPE_CHECKING
+from typing import (
+    Any,
+    Iterable,
+    Literal,
+    Optional,
+    Tuple,
+    TYPE_CHECKING,
+    Union,
+    overload,
+)
 from rqdb.errors import DBError
+from rqdb.explain import (
+    ExplainQueryPlan,
+    format_explain_query_plan_result,
+    parse_explain_query_plan,
+    print_explain_query_plan_result,
+    write_explain_query_plan,
+)
 from rqdb.logging import log
 from rqdb.result import BulkResult, ResultItem, ResultItemCursor
 from rqdb.preprocessing import get_sql_command, clean_nulls
@@ -7,7 +23,7 @@ import time
 import secrets
 
 if TYPE_CHECKING:
-    from rqdb.connection import Connection
+    from rqdb.connection import Connection, SyncWritableIO
 
 
 class Cursor:
@@ -22,7 +38,7 @@ class Cursor:
         self.connection = connection
         """The underlying connection to the rqlite cluster."""
 
-        self.read_consistency = read_consistency
+        self.read_consistency: Literal["none", "weak", "strong"] = read_consistency
         """The read consistency to use when executing queries."""
 
         self.freshness = freshness
@@ -47,7 +63,7 @@ class Cursor:
     def execute(
         self,
         operation: str,
-        parameters: Optional[tuple] = None,
+        parameters: Optional[Iterable[Any]] = None,
         raise_on_error: bool = True,
         read_consistency: Optional[Literal["none", "weak", "strong"]] = None,
         freshness: Optional[str] = None,
@@ -58,7 +74,7 @@ class Cursor:
 
         Args:
             operation (str): The query to execute.
-            parameters (tuple): The parameters to pass to the query.
+            parameters (iterable): The parameters to pass to the query.
             raise_on_error (bool): If True, raise an error if the query fails. If
                 False, you can check the result item's error property to see if
                 the query failed.
@@ -85,12 +101,12 @@ class Cursor:
 
         command = get_sql_command(operation)
         cleaned_query, parameters = clean_nulls(operation, parameters)
-        is_read = command == "SELECT"
+        is_read = command in ("SELECT", "EXPLAIN")
         request_id = secrets.token_hex(4)
 
         if is_read:
 
-            def msg_supplier(max_length: Optional[int]) -> str:
+            def msg_supplier_1(max_length: Optional[int]) -> str:
                 abridged_query = operation
                 if max_length is not None and len(abridged_query) > max_length:
                     abridged_query = abridged_query[:max_length] + "..."
@@ -105,7 +121,7 @@ class Cursor:
 
                 return f"  [RQLITE {command} @ {read_consistency}{freshness_str} {{{request_id}}}] - {repr(abridged_query)}; {abridged_parameters}"
 
-            log(self.connection.log_config.read_start, msg_supplier)
+            log(self.connection.log_config.read_start, msg_supplier_1)
 
             path = f"/db/query?level={read_consistency}"
             if read_consistency == "none":
@@ -120,7 +136,7 @@ class Cursor:
             )
         else:
 
-            def msg_supplier(max_length: Optional[int]) -> str:
+            def msg_supplier_2(max_length: Optional[int]) -> str:
                 abridged_query = operation
                 if max_length is not None and len(abridged_query) > max_length:
                     abridged_query = abridged_query[:max_length] + "..."
@@ -131,7 +147,7 @@ class Cursor:
 
                 return f"  [RQLITE {command} {{{request_id}}}] - {repr(abridged_query)}; {abridged_parameters}"
 
-            log(self.connection.log_config.write_start, msg_supplier)
+            log(self.connection.log_config.write_start, msg_supplier_2)
 
             request_started_at = time.perf_counter()
             response = self.connection.fetch_response(
@@ -144,7 +160,7 @@ class Cursor:
         payload = response.json()
         request_time = time.perf_counter() - request_started_at
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_3(max_length: Optional[int]) -> str:
             abridged_payload = response.text
             if max_length is not None and len(abridged_payload) > max_length:
                 abridged_payload = abridged_payload[:max_length] + "..."
@@ -155,7 +171,7 @@ class Cursor:
             self.connection.log_config.read_response
             if is_read
             else self.connection.log_config.write_response,
-            msg_supplier,
+            msg_supplier_3,
         )
 
         if "error" in payload:
@@ -193,8 +209,8 @@ class Cursor:
 
     def executemany2(
         self,
-        operations: Tuple[str],
-        seq_of_parameters: Optional[Tuple[Tuple[Any]]] = None,
+        operations: Iterable[str],
+        seq_of_parameters: Optional[Iterable[Iterable[Any]]] = None,
         transaction: bool = True,
         raise_on_error: bool = True,
     ) -> BulkResult:
@@ -208,8 +224,8 @@ class Cursor:
         as if they are updates, i.e., no result rows will be returned.
 
         Args:
-            operations (Tuple[str]): The operations to execute.
-            seq_of_parameters (Tuple[Tuple[Any]]): The parameters to pass to each operation.
+            operations (Iterable[str]): The operations to execute.
+            seq_of_parameters (Iterable[Iterable[Any]]): The parameters to pass to each operation.
             transaction (bool): If True, execute the operations within a transaction.
             raise_on_error (bool): If True, raise an error if any of the operations fail. If
                 False, you can check the result item's error property to see if the
@@ -224,30 +240,51 @@ class Cursor:
         if seq_of_parameters is None:
             seq_of_parameters = tuple(tuple() for _ in operations)
 
-        if len(operations) != len(seq_of_parameters):
-            raise ValueError(
-                f"Number of operations ({len(operations)}) does not match number of parameters ({len(seq_of_parameters)})"
-            )
-
         path = "/db/execute"
         if transaction:
             path += "?transaction"
 
         cleaned_request = []
-        for operation, parameters in zip(operations, seq_of_parameters):
+
+        operations_iter = iter(operations)
+        seq_of_parameters_iter = iter(seq_of_parameters)
+
+        next_operation: str = ""
+        next_parameters: Iterable[Any] = tuple()
+
+        while True:
+            try:
+                next_operation = next(operations_iter)
+            except StopIteration:
+                operations_iter = None
+            try:
+                next_parameters = next(seq_of_parameters_iter)
+            except StopIteration:
+                seq_of_parameters_iter = None
+
+            if (operations_iter is None) is not (seq_of_parameters_iter is None):
+                raise ValueError(
+                    "operations and seq_of_parameters must be the same length"
+                )
+
+            if operations_iter is None or seq_of_parameters_iter is None:
+                break
+
+            operation: str = next_operation
+            parameters: Iterable[Any] = next_parameters
             cleaned_query, parameters = clean_nulls(operation, parameters)
             cleaned_request.append([cleaned_query, *parameters])
 
         request_id = secrets.token_hex(4)
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_1(max_length: Optional[int]) -> str:
             abridged_request = repr(cleaned_request)
             if max_length is not None and len(abridged_request) > max_length:
                 abridged_request = abridged_request[:max_length] + "..."
 
             return f"  [RQLITE BULK {path} {{{request_id}}}] - {abridged_request}"
 
-        log(self.connection.log_config.write_start, msg_supplier)
+        log(self.connection.log_config.write_start, msg_supplier_1)
 
         request_started_at = time.perf_counter()
         response = self.connection.fetch_response(
@@ -259,7 +296,7 @@ class Cursor:
         payload = response.json()
         request_time = time.perf_counter() - request_started_at
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_2(max_length: Optional[int]) -> str:
             abridged_payload = response.text
             if max_length is not None and len(abridged_payload) > max_length:
                 abridged_payload = abridged_payload[:max_length] + "..."
@@ -268,7 +305,7 @@ class Cursor:
 
         log(
             self.connection.log_config.write_response,
-            msg_supplier,
+            msg_supplier_2,
         )
 
         result = BulkResult.parse(payload)
@@ -279,7 +316,7 @@ class Cursor:
 
     def executemany3(
         self,
-        operation_and_parameters: Tuple[Tuple[str, Tuple[Any]]],
+        operation_and_parameters: Iterable[Tuple[str, Iterable[Any]]],
         transaction: bool = True,
         raise_on_error: bool = True,
     ) -> BulkResult:
@@ -317,14 +354,14 @@ class Cursor:
 
         request_id = secrets.token_hex(4)
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_1(max_length: Optional[int]) -> str:
             abridged_request = repr(cleaned_request)
             if max_length is not None and len(abridged_request) > max_length:
                 abridged_request = abridged_request[:max_length] + "..."
 
             return f"  [RQLITE BULK {path} {{{request_id}}}] - {abridged_request}"
 
-        log(self.connection.log_config.write_start, msg_supplier)
+        log(self.connection.log_config.write_start, msg_supplier_1)
 
         request_started_at = time.perf_counter()
         response = self.connection.fetch_response(
@@ -336,7 +373,7 @@ class Cursor:
         payload = response.json()
         request_time = time.perf_counter() - request_started_at
 
-        def msg_supplier(max_length: Optional[int]) -> str:
+        def msg_supplier_2(max_length: Optional[int]) -> str:
             abridged_payload = response.text
             if max_length is not None and len(abridged_payload) > max_length:
                 abridged_payload = abridged_payload[:max_length] + "..."
@@ -345,7 +382,7 @@ class Cursor:
 
         log(
             self.connection.log_config.write_response,
-            msg_supplier,
+            msg_supplier_2,
         )
 
         result = BulkResult.parse(payload)
@@ -354,6 +391,144 @@ class Cursor:
 
         return result
 
+    @overload
+    def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Literal["str"] = "str",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+    ) -> str:
+        ...
+
+    @overload
+    def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Literal["print"],
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+    ) -> None:
+        ...
+
+    @overload
+    def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Literal["plan"],
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+    ) -> ExplainQueryPlan:
+        ...
+
+    @overload
+    def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: "SyncWritableIO",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+    ) -> None:
+        ...
+
+    def explain(
+        self,
+        operation: str,
+        parameters: Optional[Iterable[Any]] = None,
+        *,
+        out: Union[Literal["str", "print", "plan"], "SyncWritableIO"] = "str",
+        read_consistency: Optional[Literal["none", "weak"]] = None,
+        freshness: Optional[str] = None,
+        indent: int = 3,
+        include_raw: bool = False,
+    ) -> Union[ExplainQueryPlan, str, None]:
+        """Accepts any query; if it is not prefixed with EXPLAIN then
+        it will be prefixed with EXPLAIN QUERY PLAN. The reuslt will
+        then be parsed into the corresponding tree structure and either
+        written to the designated output or returned.
+
+        This always raises if there is an error, and can only operate
+        at none/weak level. If the read consistency is not specified
+        and the cursor read consistency is strong, weak consistency
+        is used instead.
+
+        Args:
+            operation (str): The query to execute.
+            out (Union[Literal["str", "print", "plan"], SyncWritableIO]):
+                The output to write the explain query plan to. If "str",
+                return the explain query plan as a string. If "print",
+                print the explain query plan to stdout. If "plan", return
+                the explain query plan as a tree structure. If a writable
+                stream, write the explain query plan to the stream.
+            parameters (Optional[Iterable[Any]]): The parameters to pass to the query.
+                Parameters must be specified if the operation is a parameterized query,
+                though the values generally only need to be of the correct shape for
+                the appropriate plan to be returned
+            read_consistency ("none", "weak", None): The read consistency to use,
+                None for a random node, weak for the current leader, or None for the
+                cursor default (downgrading strong to weak as query plans are not
+                necessarily deterministic)
+            freshness (Optional[str]): The freshness to use when executing
+                none read consistency queries. If None, use the default freshness
+                for this cursor.
+            indent (int): The number of characters to indent each level when
+                formatting the plan. Ignored for "plan" output.
+            include_raw (bool): If True, include the ids that made each row in the
+                formatted output. Ignored for "plan" output.
+
+        Returns:
+            Depends on the value of out. If "str", return the explain query plan as a
+            string. If "print", print the explain query plan to stdout. If "plan",
+            return the explain query plan as a tree structure. If a writable stream,
+            write the explain query plan to the bytes stream and return None
+        """
+        command = get_sql_command(operation)
+        if command != "EXPLAIN":
+            operation = f"EXPLAIN QUERY PLAN {operation}"
+
+        if read_consistency is None:
+            read_consistency = (
+                self.read_consistency if self.read_consistency != "strong" else "weak"
+            )
+
+        result = self.execute(
+            operation,
+            parameters,
+            read_consistency=read_consistency,
+            freshness=freshness,
+        )
+        if out == "plan":
+            return parse_explain_query_plan(result)
+        elif out == "str":
+            return format_explain_query_plan_result(
+                result, indent=indent, include_raw=include_raw
+            )
+        elif out == "print":
+            print_explain_query_plan_result(
+                result, indent=indent, include_raw=include_raw
+            )
+            return
+        else:
+            assert not isinstance(out, str), f"unrecognized out value: {out}"
+
+            plan = parse_explain_query_plan(result)
+            write_explain_query_plan(plan, out, indent=indent, include_raw=include_raw)
+            return
+
     def fetchone(self) -> Optional[list]:
         """Fetches the next row from the cursor. If there are no more rows,
         returns None.
